""" 
.. inheritance-diagram:: pyopus.optimizer.vptree
    :parts: 1

**Vantage Point-tree (VP-tree) implementation (PyOPUS subsystem name: VPT)**

Based on VPTree from PyPI. 
"""
from ..misc.debug import DbgMsgOut, DbgMsg

import numpy as np
import bisect

__all__ = [ "VPTree" ]


class VPTree:
    """ 
    VP-Tree data structure for efficient nearest neighbour search.

    The VP-tree is a data structure for efficient nearest neighbour
    searching and finds the nearest neighbour in O(log n)
    complexity given a tree constructed of n data points. Construction
    complexity is O(n log n).

    Parameters
    ----------
    points : Iterable
        Construction points.
    dist_fn : Callable
        Function taking point instances as arguments and returning
        the distance between them.
    auto_rebuild_depth_factor : int
        Amount of excess tree depth generated by calls to :meth:`append` 
        that results in a tree rebuild. 
    unique : boolean
        Require points in the tree to be unique
    root : boolean (internal argument)
        When ``True`` the tree is built from supplied points, 
        otherwirse a blank root node is created
    """

    def __init__(self, points, dist_fn, auto_rebuild_depth_factor=100, unique=True, root=True):
        # Left and right tree
        self.left = None
        self.right = None
        # Range of distances where left tree points are located
        self.left_min = np.inf
        self.left_max = 0
        # Range of distances where right tree points are located
        self.right_min = np.inf
        self.right_max = 0
        
        # Root node
        if root:
            # Number of points
            self.n_points=0
            
            # Function for computing distance 
            if dist_fn is None:
                raise ValueError(DbgMsg("VPT", "Need a distance function."))
            self.dist_fn = dist_fn
            
            # Unique entries required
            self.unique=unique
            
            # Store auto rebuild tree depth factor 
            if auto_rebuild_depth_factor<=1:
                raise ValueError(DbgMsg("VPT", "Automatic rebuild tree depth factor must be greater than 1"))
            self.auto_rebuild_depth_factor=auto_rebuild_depth_factor
            
            # Build if at least one point is given
            if points is not None and len(points)>0:
                # Build tree and store tree depth in root node
                self.tree_depth=self._build(points)
                self.n_points=len(points)
            else:
                self.tree_depth=0
            
    def points(self):
        """
        Return points in the tree as a list. 
        """
        l=list()
        nodes_to_visit=[self]
        while len(nodes_to_visit)>0:
            node=nodes_to_visit.pop()
            l.append(node.vp)
            if node.left is not None:
                nodes_to_visit.append(node.left)
            if node.right is not None:
                nodes_to_visit.append(node.right)
        
        return l
    
    def depth(self):
        """
        Returns the depth of the tree. 
        """
        return self.tree_depth
    
    def npoints(self):
        """
        Returns the number of points in the tree.  
        """
        return self.n_points
    
    def _split(self, points):
        """
        Private method. Splits a set equally with respect to the distances 
        from the vantage point into inner and outer points. 
        
        Parameters
        ----------
        points : Iterable
            Construction points.
        
        First point in *points* is chosen as the vantage point. 
        
        Returns a tuple:
          
        * vantage_point
        * left_min (inf by default) - distance to the closest inner point
        * left_max (0 by default) - distance to the farthest inner point
        * left_points - list of inner points
        * right_min (inf by default) - distance to the closest outer point
        * right_max (0 by default) - distance to the farthest outer point
        * right_points - list of outer points
        """
        # Choose vantage point (first point), remove it from list
        vp=points[0]
        points=points[1:]
        
        left_points = []
        right_points = []
        left_min = np.inf
        left_max = 0
        right_min = np.inf
        right_max = 0
        
        # Split points in left (within median distance) and right (outside median distance)
        if len(points)>0:
            # Compute distances of all points to the vantage point
            distances = [self.dist_fn(vp, p) for p in points]
            
            # Compute median distance ()
            median = np.median(distances)
            
            # Scan points, sort them in two sets
            for point, distance in zip(points, distances):
                # This will put two points with the same distance from vp into the same subset
                # Eventually one of them will become the vp while the other will have zero distance from the vp. 
                # At that point we detect a duplicate. 
                if self.unique and distance==0.0:
                    # Duplicate detected, throw it out
                    continue
                
                if distance >= median:
                    # Update min and max distance from vantage point
                    right_min = min(distance, right_min)
                    if distance > right_max:
                        right_max = distance
                        right_points.insert(0, point) # put furthest first
                    else:
                        # If point is not the furthest point, add it to the end of the list
                        right_points.append(point)
                else:
                    # Update min and max distance from vantage point
                    left_min = min(distance, left_min)
                    if distance > left_max:
                        left_max = distance
                        left_points.insert(0, point) # put furthest first
                    else:
                        # If point is not the furthest point, add it to the end of the list
                        left_points.append(point)
        
        return (vp, left_min, left_max, left_points, right_min, right_max, right_points)
        
    def _build(self, points):
        """
        Private method. Builds a VP tree in a blank root node. 
        Chooses a VP for the node and builds the two subtrees. 
        
        Parameters
        ----------
        points : Iterable
            Construction points.
        
        No recursive calls used. 
        
        Returns tree depth. 
        """
        if len(points)==0:
            return 0
        
        # Stack holding split sets that need to be added to the tree
        # Entry: node, split_set, depth
        sstack=list()
        max_depth=1
        
        # Prepare initial split
        split_set=self._split(points)
        sstack.append((self, split_set, 1))
        
        while len(sstack)>0:
            node, (vp, left_min, left_max, left_points, right_min, right_max, right_points), depth = sstack.pop()
            
            # Update max depth
            if depth>max_depth:
                max_depth=depth
            
            # Record vantage point and ranges
            node.vp=vp
            node.left_min=left_min
            node.left_max=left_max
            node.right_min=right_min
            node.right_max=right_max
            
            # Left subtree
            if len(left_points)>0:
                # Subtree
                node.left=VPTree(None, None, root=False)
                sstack.append((node.left, self._split(left_points), depth+1))
            
            # Right subtree
            if len(right_points)>0:
                # Subtree
                node.right=VPTree(None, None, root=False)
                sstack.append((node.right, self._split(right_points), depth+1))
        
        return max_depth
        
    def rebuild(self, dist_fn=None):
        """
        Rebuilds the tree. 
        
        Parameters
        ----------
        dist_fn : Callable
            Function taking point instances as arguments and returning
            the distance between them. Optional. If specified the tree
            is erbuild according to the specified dist_fn. 
    
        A rebuild reduces the tree depth if any points were appended to the 
        tree after it was built/rebuild. 
        """
        if self.n_points==0:
            return
        
        # Collect all points
        points=self.points()
        
        # Reset root
        self.left = None
        self.right = None
        self.left_min = np.inf
        self.left_max = 0
        self.right_min = np.inf
        self.right_max = 0
        
        # Optional new distance function
        if dist_fn is not None:
            self.dist_fn=dist_fn
            
        # Rebuild
        self.tree_depth=self._build(points)
    
    def append(self, p):
        """
        Append a point. Updates point counts, min/max ranges, and depth. 
        
        Parameters
        ----------
        p : Point to add. 
        
        Returns ``None`` if a point is appended. 
        
        If the point is a duplicate and *unique* was set to `True` at tree 
        construction, returns the stored point from the tree. 
        """
        # Is this the first point?
        if self.n_points==0:
            # Add it as VP, leave min/max data unchanged, no subtrees
            self.vp=p
            self.tree_depth=1
            self.n_points+=1
            return True
        
        # Look for node where the point should be appended, update min/max and point counts
        # Start at root node which lies at depth 1
        node=self
        depth=1
        while True:
            # Compute distance of point from VP
            d=self.dist_fn(node.vp, p)
            
            # Is this point a duplicate? 
            if d==0 and self.unique:
                # Yes, stop search and return the stored entry (point is a duplicate)
                return node.vp
            
            # Check point position with respect to the inner/outer belt
            if d-node.left_max<=node.right_min-d:
                # If p lies closer to the inner belt than to the outer belt, add it to the inner tree (left)
                # Update belt boundaries
                dmax=node.left_max
                dmin=node.left_min
                if d>node.left_max:
                    node.left_max=d
                if d<node.left_min:
                    node.left_min=d
                # Do we have a left tree
                if node.left is None:
                    # No, build leaf node
                    node.left=VPTree(points=None, dist_fn=None, root=False)
                    node.left._build(points=[p])
                    # Increase depth and exit loop
                    depth+=1
                    #print("ap left", dmin, dmax)
                    break
                else:
                    # Else, continue search in the left tree, increase depth
                    node=node.left
                    depth+=1
            else:
                # Else, add it to the outer tree (right). Points in the middle are added to the outer tree. 
                # Update belt boundaries
                if d>node.right_max:
                    node.right_max=d
                if d<node.right_min:
                    node.right_min=d
                # Do we have a right tree
                if node.right is None:
                    # No, build leaf node
                    node.right=VPTree(points=None, dist_fn=None, root=False)
                    node.right._build(points=[p])
                    # Increase depth and exit loop
                    depth+=1
                    # print("ap right")
                    break
                else:
                    # Else, continue search in the right tree, increase depth
                    node=node.right
                    depth+=1
            
            # Update tree depth
            if depth>self.tree_depth:
                self.tree_depth=depth
                
        # Compute expected tree depth, rebuild if necessary
        expected_depth=np.ceil(np.log(self.n_points)/np.log(2))
        if self.tree_depth>expected_depth*self.auto_rebuild_depth_factor:
            print("automatic rebuild at n=", self.n_points, ", depth=", self.tree_depth, ", expected depth=", expected_depth)
            self.rebuild()
            # print("rebuilt", self.tree_depth, expected_depth*self.auto_rebuild_depth_factor, self.n_points)
        
        # Point added
        self.n_points+=1
        return None
    
    def get_nearest_neighbour(self, query):
        """ 
        Get single nearest neighbour.
        
        Parameters
        ----------
        query : Any
            Query point.

        Returns
        -------
        Any
            Single nearest neighbour.
            A neighbour is a tuple (distance, point). 
            If no neighbour is found, return None. 
        """
        ret=self.get_n_nearest_neighbours(query, n_neighbours=1)
        if len(ret)==0:
            return None
        else:
            return ret[0]

    def get_n_nearest_neighbours(self, query, n_neighbours):
        """ 
        Get `n_neighbours` nearest neigbours to `query`
        
        Parameters
        ----------
        query : Any
            Query point.
        n_neighbours : int
            Number of neighbours to fetch.

        Returns
        -------
        list
            List of `n_neighbours` nearest neighbours.
            A neighbour is a tuple (distance, point). 
        """
        if not isinstance(n_neighbours, int) or n_neighbours < 1:
            raise ValueError(DbgMsg("VPT", 'n_neighbours must be strictly positive integer'))
        
        # List of closest neighbours
        neighbours = _AutoSortingList(max_size=n_neighbours)
        
        if self.n_points==0:
            return []
        
        # (node, distance from query point to min/max belt)
        # Use 0 for this distance so that this node gets examined no matter where the query point lies
        nodes_to_visit = [(self, 0)]

        # Distance of furthest neighbour from query point
        furthest_d = np.inf
        
        # While we have nodes to visit
        while len(nodes_to_visit) > 0:
            node, d0 = nodes_to_visit.pop(0)
            
            # Skip this node if distance from min/max belt is greater than furthest_d
            if d0 > furthest_d:
                continue
            
            # Compute distance of node's VP to query point
            d = self.dist_fn(query, node.vp)
            # If VP is within furtest_d, add it to the list of neighbours, update furthest_d
            if d < furthest_d:
                neighbours.append((d, node.vp))
                # Get last neighbour (the one that is furthest away), set furthest_d to its distance
                furthest_d, _ = neighbours[-1]
            
            # Do we have a left tree
            if node.left is not None:
                if node.left_min <= d <= node.left_max:
                    # Ball B(query, furthest_d) is centered within the inner min/max belt
                    # Add left tree to search, belt distance is 0
                    nodes_to_visit.insert(0, (node.left, 0))
                elif node.left_min - furthest_d <= d <= node.left_max + furthest_d:
                    # Ball B(query, furtest_d) intersects the inner min/max belt
                    # Add left tree to search with belt distance
                    #   min-d if d<min (query point inside B(vp, min)); this is always positive
                    #   d-max if d>min (query point outside B(vp, min) or on its surface); this distance can also be negative
                    nodes_to_visit.append((node.left,  node.left_min - d if d < node.left_min else d - node.left_max))
            
            # Do we have a right tree
            if node.right is not None:
                if node.right_min <= d <= node.right_max:
                    # Ball B(query, furtest_d) is centered within the outer min/max belt
                    # Add right tree to search, belt distance is 0
                    nodes_to_visit.insert(0, (node.right, 0))
                elif node.right_min - furthest_d <= d <= node.right_max + furthest_d:
                    # Ball B(query, furtest_d) intersects the outer min/max belt
                    # Add right tree to search with belt distance
                    #   min-d if d<min (query point inside B(vp, min)); this is always positive
                    #   d-max if d>min (query point outside B(vp, min) or on its surface); this distance can also be negative
                    nodes_to_visit.append((node.right, node.right_min - d if d < node.right_min else d - node.right_max))

        return list(neighbours)

    def get_all_in_range(self, query, max_distance):
        """ Find all neighbours within `max_distance`.

        Parameters
        ----------
        query : Any
            Query point.
        max_distance : float
            Threshold distance for query.

        Returns
        -------
        neighbours : list
            List of all matches within `max_distance`.
            A match is a tuple (distance, point).

        Notes
        -----
        Returned neighbours are not sorted according to distance.
        """
        neighbours = list()
        # (node, distance from query point to min/max belt)
        # Use 0 for this distance so that this node gets examined no matter where the query point lies
        nodes_to_visit = [(self, 0)]
        
        # While we have nodes to visit
        while len(nodes_to_visit) > 0:
            node, d0 = nodes_to_visit.pop(0)
            # Skip this node if distance from min/max belt is greater than max_distance
            if d0 > max_distance:
                continue
            
            # Compute distance of node's VP to query point
            d = self.dist_fn(query, node.vp)
            # If VP is within max_distance, add it to the list of neighbours
            if d < max_distance:
                neighbours.append((d, node.vp))

            # Do we have a left tree
            if node.left is not None:
                if node.left_min <= d <= node.left_max:
                    # Ball B(query, max_distance) is centered within the inner min/max belt
                    # Add left tree to search, belt distance is 0
                    nodes_to_visit.insert(0, (node.left, 0))
                elif node.left_min - max_distance <= d <= node.left_max + max_distance:
                    # Ball B(query, max_distance) intersects the inner min/max belt
                    # Add left tree to search with belt distance
                    #   min-d if d<min (query point inside B(vp, min)); this is always positive
                    #   d-max if d>min (query point outside B(vp, min) or on its surface); this distance can also be negative
                    nodes_to_visit.append(
                        (node.left, node.left_min - d if d < node.left_min else d - node.left_max))
            
            # Do we have a right tree
            if node.right is not None:
                if node.right_min <= d <= node.right_max:
                    # Ball B(query, max_distance) is centered within the outer min/max belt
                    # Add right tree to search, belt distance is 0
                    nodes_to_visit.insert(0, (node.right, 0))
                elif node.right_min - max_distance <= d <= node.right_max + max_distance:
                    # Ball B(query, max_distance) intersects the outer min/max belt
                    # Add right tree to search with belt distance
                    #   min-d if d<min (query point inside B(vp, min)); this is always positive
                    #   d-max if d>min (query point outside B(vp, min) or on its surface); this distance can also be negative
                    nodes_to_visit.append((node.right, node.right_min - d if d < node.right_min else d - node.right_max))

        return neighbours


class _AutoSortingList(list):
    """ 
    Simple auto-sorting list (lexically ascending). 
    
    Parameters
    ---------
    max_size : int, optional
        Max queue size.
    """

    def __init__(self, max_size=None, *args):
        super(_AutoSortingList, self).__init__(*args)
        self.max_size = max_size

    def append(self, item):
        """ Append `item` and sort.

        Parameters
        ----------
        item : Any
            Input item.
        """
        bisect.insort_right(self, item)
        if len(self)>self.max_size:
            self.pop()
        

if __name__=="__main__":
    import time
    from pprint import pprint
    import cProfile
    
    def main():
        # Euclidean distance
        def euclidean(p, q):
            return ((p-q)**2).sum()**0.5
        
        # Slow search
        def slow_get_all_in_range(points, q, r, dist_fn):
            results=list()
            for point in points:
                d=dist_fn(q, point)
                if d<r:
                    results.append((d, point))
        
            return results

        # Dimension
        n=5
        
        # Point count
        N=31000
        
        # Seed
        np.random.seed(0)
        
        # Generate random points
        points=np.random.rand(N,n)
        
        ## Create VP tree
        print("Construct from scratch")
        t0=time.perf_counter()
        tree=VPTree(points, euclidean)
        dt=time.perf_counter()-t0
        print("Construction time:", dt)
        print("Depth:", tree.depth(), "Count:", tree.npoints(), "\n")
        
        # # Rebuild
        # print("Rebuild")
        # t0=time.perf_counter()
        # tree.rebuild()
        # dt=time.perf_counter()-t0
        # print("Rebuild time:", dt)
        # print("Depth:", tree.depth(), "Count:", tree.npoints(), "\n")
        # 
        # # Build with appending
        # print("Build by appending")
        # t0=time.perf_counter()
        # # Blank tree
        # tree=VPTree([], euclidean, auto_rebuild_depth_factor=2.5)
        # # Append points
        # for p in points:
        #     tree.append(p)
        # dt=time.perf_counter()-t0
        # print("Append build time:", dt)
        # print("Depth:", tree.depth(), "Count:", tree.npoints(), "\n")
        # 
        # Try appending a duplicate
        # print("Append a duplicate")
        # n1=tree.npoints()
        # ret=tree.append(points[0])
        # print("Initial points", n1, ", new points", tree.npoints())
        # print("Found entry:", ret, "\n")
        # 
        # # Try appending a new point
        # print("Append a new point")
        # n1=tree.npoints()
        # ret=tree.append(np.ones(n))
        # print("Initial points", n1, ", new points", tree.npoints())
        # print("Found entry:", ret, "\n")
        
        # Find all points within dist of origin
        print("Find within radius")
        q=np.zeros(n)
        dist=0.3
        t0=time.perf_counter()
        found=tree.get_all_in_range(q, dist)
        dt=time.perf_counter()-t0
        print("Found %d points within %f of origin" % (len(found), dist))
        print("Search time:", dt, "\n")
        
        # Sort according to distance, vector componets (lexically)
        found.sort()
        
        # Build result matrix
        V1=np.array([ p for _, p in found ])
        
        # Slow search
        # Find all points within 0.1 from origin
        print("Find within radius (brute force)")
        q=np.zeros(n)
        t0=time.perf_counter()
        found=slow_get_all_in_range(points, q, dist, euclidean)
        dt=time.perf_counter()-t0
        print("Search time:", dt)
        
        # Sort according to distance, vector componets (lexically)
        found.sort()
        
        # Build result matrix
        V2=np.array([ p for _, p in found ])
        
        # Print difference
        print("Point differences between tree and brute force search")
        pprint(V1-V2)
        print()
        
        # # Find closest 10 points to the origin
        # print("Find within radius")
        # q=np.zeros(n)
        # t0=time.perf_counter()
        # found=tree.get_n_nearest_neighbours(q, 10)
        # dt=time.perf_counter()-t0
        # print("Found %d points closest to the origin" % (len(found)))
        # print("Search time:", dt, "\n")
        # pprint(found)
        
        # Sort according to distance, vector componets (lexically)
        found.sort()
    
    # cProfile.run('main()')
    main()
    
